import time
import sys
import pprint
import unittest
import subprocess
import tempfile
import datetime
from threading import Timer
import shutil
import logging
from orthanc_api_client import OrthancApiClient, ChangeType
from orthanc_api_client import helpers
import orthanc_api_client.exceptions as api_exceptions
import pathlib
import os
import logging

from orthanc_tools import OrthancCloner, ClonerMode, OrthancMonitor, OrthancTestDbPopulator, PacsMigrator, OrthancComparator, OrthancForwarder, ForwarderMode, ForwarderDestination, OrthancCleaner

here = pathlib.Path(__file__).parent.resolve()

logger = logging.getLogger('orthanc_tools')
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)


class Test3Orthancs(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        subprocess.run(["docker-compose", "down", "-v"], cwd=here/"docker-setup")
        subprocess.run(["docker-compose", "up", "-d"], cwd=here/"docker-setup")

        cls.oa = OrthancApiClient('http://localhost:10042', user='test', pwd='test')
        cls.oa.wait_started()
        cls.ob = OrthancApiClient('http://localhost:10043', user='test', pwd='test')
        cls.ob.wait_started()
        cls.oc = OrthancApiClient('http://localhost:10044', user='test', pwd='test')
        cls.oc.wait_started()

    @classmethod
    def tearDownClass(cls):
        subprocess.run(["docker-compose", "down", "-v"], cwd=here/"docker-setup")

    def test_cloner_default(self):
        self.oa.delete_all_content()
        self.ob.delete_all_content()

        self.oa.upload_file(here / "stimuli/CT_small.dcm")

        cloner = OrthancCloner(source=self.oa, destination=self.ob)
        cloner.execute()

        self.assertEqual(len(self.oa.instances.get_all_ids()), len(self.ob.instances.get_all_ids()))

    def test_cloner_transfer(self):
        self.oa.delete_all_content()
        self.ob.delete_all_content()

        self.oa.upload_file(here / "stimuli/CT_small.dcm")

        cloner = OrthancCloner(source=self.oa, destination_peer='orthanc-b', mode=ClonerMode.TRANSFER)
        time.sleep(5) # wait for StableStudy event
        cloner.execute()

        self.assertEqual(len(self.oa.instances.get_all_ids()), len(self.ob.instances.get_all_ids()))

    def test_cloner_peering(self):
        self.oa.delete_all_content()
        self.ob.delete_all_content()

        self.oa.upload_file(here / "stimuli/CT_small.dcm")

        cloner = OrthancCloner(source=self.oa, destination_peer='orthanc-b', mode=ClonerMode.PEERING)
        cloner.execute()

        self.assertEqual(len(self.oa.instances.get_all_ids()), len(self.ob.instances.get_all_ids()))

    def test_monitor(self):
        self.oa.delete_all_content()
        processed_instances = []

        monitor = OrthancMonitor(
            self.oa,
            polling_interval=0.1
        )

        def new_instance_handler(change_id, instance_id, api_client):
            processed_instances.append(instance_id),

        monitor.add_handler(ChangeType.NEW_INSTANCE, new_instance_handler)

        monitor.start()

        self.oa.upload_file(here / "stimuli/CT_small.dcm")

        helpers.wait_until(lambda: len(processed_instances) > 0, 30)

        monitor.stop()
        self.assertEqual(1, len(processed_instances))

    def test_populator_repeatability(self):
        self.oa.delete_all_content()
        self.ob.delete_all_content()

        populator_a = OrthancTestDbPopulator(api_client=self.oa, studies_count=2, random_seed=42)
        populator_b = OrthancTestDbPopulator(api_client=self.ob, studies_count=2, random_seed=42)

        populator_a.execute()
        populator_b.execute()

        self.assertEqual(self.oa.instances.get_all_ids(), self.ob.instances.get_all_ids())

    def test_populator_repeatability_with_forced_quantities(self):
        self.oa.delete_all_content()
        self.ob.delete_all_content()

        populator_a = OrthancTestDbPopulator(api_client=self.oa, studies_count=2, random_seed=42, series_count=1, instances_count=1)
        populator_b = OrthancTestDbPopulator(api_client=self.ob, studies_count=2, random_seed=42, series_count=1, instances_count=1)

        populator_a.execute()
        populator_b.execute()

        a_ids = self.oa.instances.get_all_ids()
        b_ids = self.ob.instances.get_all_ids()
        self.assertEqual(a_ids, b_ids)
        self.assertEqual(len(a_ids), 2)


    def test_monitor_recovery(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            persist_status_path = os.path.join(temp_dir, 'seq.txt')

            processed_resources = []

            def process_instance(change_id, instance_id, api_client):
                logger.info(f'processing instance {instance_id}')
                time.sleep(5)
                processed_resources.append(instance_id)

            def process_series(change_id, series_id, api_client):
                logger.info(f'processing series {series_id}')
                processed_resources.append(series_id)

            monitor = OrthancMonitor(
                self.oa,
                polling_interval=0.1,
                persist_status_path=persist_status_path,
                worker_threads_count=4
            )

            # first event is lengthy (5 seconds) and will not be processed at the time we first check the sequence id file
            monitor.add_handler(ChangeType.NEW_INSTANCE, process_instance)
            monitor.add_handler(ChangeType.NEW_SERIES, process_series)

            self.oa.upload_file(here / "stimuli/CT_small.dcm")

            monitor.start()
            helpers.wait_until(lambda: len(processed_resources) > 0, 1)
            helpers.wait_until(lambda: os.path.exists(persist_status_path), 1)
            with open(persist_status_path, "rt") as f:
                seq_id = int(f.read())

            all_changes, last_seq_id, done = self.oa.get_changes()
            all_series_ids = self.oa.series.get_all_ids()
            all_instances_ids = self.oa.instances.get_all_ids()

            self.assertGreaterEqual(all_changes[0].sequence_id, seq_id)        # change '1' has not been processed yet
            self.assertEqual(all_series_ids[0], processed_resources[0])     # change '2' has been processed
            self.assertEqual(1, len(processed_resources))  # the instance has not been processed yet because of the sleep 5

            monitor.stop()

            helpers.wait_until(lambda: len(processed_resources) == 2, 6)
            self.assertEqual(2, len(processed_resources))  # the instance should have been processed by now

            with open(persist_status_path, "rt") as f:
                seq_id = int(f.read())

            self.assertLessEqual(all_changes[-1].sequence_id, seq_id)        # all changes have been processed

    def test_pacs_migrator_orthanc_as_source(self):
        self.oa.delete_all_content()  # source & migrator
        self.ob.delete_all_content()  # destination

        self.oa.upload_file(here / "stimuli/CT_small.dcm")

        migrator = PacsMigrator(
            api_client=self.oa,
            destination_modality="orthanc-b",
            from_study_date=datetime.date(2004, 1, 18),
            to_study_date=datetime.date(2004, 1, 20),
            delete_from_source=False
        )
        migrator.execute()

        # check all instances have been transferred and are still on the source
        self.assertEqual(len(self.oa.instances.get_all_ids()), len(self.ob.instances.get_all_ids()))

    def test_pacs_migrator_orthanc_as_source_reverse(self):
        self.oa.delete_all_content()  # source & migrator
        self.ob.delete_all_content()  # destination

        self.oa.upload_file(here / "stimuli/CT_small.dcm")

        migrator = PacsMigrator(
            api_client=self.oa,
            destination_modality="orthanc-b",
            from_study_date=datetime.date(2004, 1, 20),
            to_study_date=datetime.date(2004, 1, 18),
            delete_from_source=False
        )
        migrator.execute()

        # check all instances have been transferred and are still on the source
        self.assertEqual(len(self.oa.instances.get_all_ids()), len(self.ob.instances.get_all_ids()))


    def test_pacs_migrator_orthanc_as_source_delete_from_source(self):
        self.oa.delete_all_content()  # source & migrator
        self.ob.delete_all_content()  # destination

        populator_a = OrthancTestDbPopulator(
            api_client=self.oa,
            studies_count=5,
            random_seed=42,
            from_study_date=datetime.date(2022, 4, 19),
            to_study_date=datetime.date(2022, 4, 25)
        )
        populator_a.execute()

        init_instances_ids = self.oa.instances.get_all_ids()

        migrator = PacsMigrator(
            api_client=self.oa,
            destination_modality="orthanc-b",
            from_study_date=datetime.date(2022, 4, 19),
            to_study_date=datetime.date(2022, 4, 25),
            delete_from_source=True
        )
        migrator.execute()

        # check all instances have been transferred and have been deleted from the source
        self.assertEqual(len(init_instances_ids), len(self.ob.instances.get_all_ids()))
        self.assertEqual(0, len(self.oa.instances.get_all_ids()))


    def test_pacs_migrator_aside(self):
        self.oa.delete_all_content()  # source
        self.ob.delete_all_content()  # migrator
        self.oc.delete_all_content()  # destination

        self.oa.upload_file(here / "stimuli/CT_small.dcm")

        migrator = PacsMigrator(
            api_client=self.ob,
            source_modality="orthanc-a",
            destination_aet="ORTHANC-C",
            from_study_date=datetime.date(2004, 1, 18),
            to_study_date=datetime.date(2004, 1, 20)
        )
        migrator.execute()

        # check all instances have been transferred and are still on the source
        self.assertEqual(len(self.oa.instances.get_all_ids()), len(self.oc.instances.get_all_ids()))

    def test_pacs_migrator_as_destination(self):
        self.oa.delete_all_content()  # source
        self.ob.delete_all_content()  # migrator & destination

        self.oa.upload_file(here / "stimuli/CT_small.dcm")

        migrator = PacsMigrator(
            api_client=self.ob,
            source_modality="orthanc-a",
            destination_aet="ORTHANC-B",
            from_study_date=datetime.date(2004, 1, 18),
            to_study_date=datetime.date(2004, 1, 20)
        )
        migrator.execute()

        # check all instances have been transferred and are still on the source
        self.assertEqual(len(self.oa.instances.get_all_ids()), len(self.ob.instances.get_all_ids()))

    def test_orthanc_comparator_as_a_migrator(self):
        self.oa.delete_all_content()  # source & migrator
        self.ob.delete_all_content()  # destination

        # populate Orthanc A & B with 2 DBs
        populator_a = OrthancTestDbPopulator(
            api_client=self.oa,
            studies_count=5,
            from_study_date=datetime.date(2022, 4, 19),
            to_study_date=datetime.date(2022, 4, 25)
        )
        populator_a.execute()

        populator_b = OrthancTestDbPopulator(
            api_client=self.ob,
            studies_count=5,
            from_study_date=datetime.date(2022, 4, 19),
            to_study_date=datetime.date(2022, 4, 25)
        )
        populator_b.execute()

        # transfer a few instances randomly before launching the comparator to make sur it acts correctly on incomplete series ...
        instances_a = self.oa.instances.get_all_ids()
        instances_b = self.ob.instances.get_all_ids()
        self.oa.modalities.send('orthanc-b', [instances_a[0], instances_a[2], instances_a[4]])
        self.ob.modalities.send('orthanc-a', [instances_b[0], instances_b[2], instances_b[4]])

        # run the comparator with B as the modality and make sure
        # everything in A goes to B
        comparator = OrthancComparator(
            api_client=self.oa,
            modality='orthanc-b',
            level='Instance',
            from_study_date=datetime.date(2022, 4, 19),
            to_study_date=datetime.date(2022, 4, 25),
            ignore_missing_from_orthanc=True,
            transfer_missing_to_modality=True
        )
        comparator.execute()

        # B should have both studies from A & B while A should stay untouched (except for the few instances transferred)
        self.assertEqual(10, len(self.ob.studies.get_all_ids()))
        self.assertNotEqual(10, len(self.oa.studies.get_all_ids()))

        # run the comparator with B as the modality and make sure
        # everything in B goes to A
        comparator = OrthancComparator(
            api_client=self.oa,
            modality='orthanc-b',
            level='Instance',
            from_study_date=datetime.date(2022, 4, 19),
            to_study_date=datetime.date(2022, 4, 25),
            ignore_missing_on_modality=True,
            retrieve_missing_from_orthanc=True
        )
        comparator.execute()

        # now both orthanc should have full dataset
        self.assertEqual(len(self.oa.instances.get_all_ids()), len(self.ob.instances.get_all_ids()))

    def test_all_instance_forwarder_modes(self):
        # simple forwarder test without filtering or processing

        for mode in [ForwarderMode.DICOM, ForwarderMode.DICOM_SERIES_BY_SERIES, ForwarderMode.DICOM_WEB, ForwarderMode.DICOM_WEB_SERIES_BY_SERIES, ForwarderMode.PEERING, ForwarderMode.TRANSFER]:
            for trigger in [ChangeType.STABLE_STUDY, ChangeType.STABLE_SERIES, ChangeType.NEW_INSTANCE]:
                self.ob.delete_all_content()  # destination
                self.oa.delete_all_content()  # source

                with OrthancForwarder(source=self.oa,
                                      destinations=[ForwarderDestination(destination="orthanc-b", forwarder_mode=mode)],
                                      trigger=trigger,
                                      polling_interval_in_seconds=0.1) as forwarder:

                    # upload once the forwarder is running
                    instances_ids = self.oa.upload_folder(here / "stimuli/MR/Brain")

                    # wait until the the source is empty (= the forwarder has completed its job)
                    helpers.wait_until(lambda: len(self.oa.studies.get_all_ids()) == 0, timeout=30)

                    self.assertEqual(len(instances_ids), len(self.ob.instances.get_all_ids()))
                    # check it has been removed from Orthanc A
                    self.assertEqual(0, len(self.oa.instances.get_all_ids()))

    def test_forwarder_when_orthanc_not_empty_at_startup(self):
        # simple forwarder test without filtering or processing

        for mode in [ForwarderMode.DICOM]:
            for trigger in [ChangeType.STABLE_STUDY, ChangeType.STABLE_SERIES, ChangeType.NEW_INSTANCE]:
                self.ob.delete_all_content()  # destination
                self.oa.delete_all_content()  # source

                # upload before the forwarder is running
                instances_ids = self.oa.upload_folder(here / "stimuli/MR/Brain")

                with OrthancForwarder(source=self.oa,
                                      destinations=[ForwarderDestination(destination="orthanc-b", forwarder_mode=mode)],
                                      trigger=trigger,
                                      polling_interval_in_seconds=0.1) as forwarder:

                    # wait until the the source is empty (= the forwarder has completed its job)
                    helpers.wait_until(lambda: len(self.oa.studies.get_all_ids()) == 0, timeout=30)

                    self.assertEqual(len(instances_ids), len(self.ob.instances.get_all_ids()))
                    # check it has been removed from Orthanc A
                    self.assertEqual(0, len(self.oa.instances.get_all_ids()))


    def test_orthanc_forwarder_filter_and_process(self):
        self.ob.delete_all_content()  # destination
        self.oa.delete_all_content()  # source

        mode = ForwarderMode.DICOM_SERIES_BY_SERIES
        trigger = ChangeType.STABLE_STUDY


        instances_ids = self.oa.upload_folder(here / "stimuli/MR/Brain")

        # keep only the sT2W/FLAIR series, delete other series
        def filter_instance(api_client, instance_id) -> bool:
            return api_client.instances.get(instance_id).series.main_dicom_tags.get('SeriesDescription') == 'sT2W/FLAIR'

        def process_instance(api_client, instance_id):
            modified = api_client.instances.modify(
                instance_id,
                replace_tags={"InstitutionName": "MY", "OtherPatientIDs": "1234"},
                keep_tags=['SOPInstanceUID', 'SeriesInstanceUID', 'StudyInstanceUID'],
                force=True,
            )
            r = api_client.upload(buffer=modified)
            self.assertEqual(r[0], instance_id)

        # Tell the target to reject incoming instances.  Therefore, we will exercise the retries !
        orthanc_a_config = self.ob.modalities.get_configuration(modality='orthanc-a')
        self.ob.modalities.delete(modality='orthanc-a')

        OrthancForwarder.retry_intervals = [1, 2, 3, 4, 5, 6]

        with OrthancForwarder(
            source=self.oa,
            destinations=[ForwarderDestination(destination="orthanc-b", forwarder_mode=mode)],
            trigger=trigger,
            polling_interval_in_seconds=0.1,
            instance_filter=filter_instance,
            instance_processor=process_instance
            ) as forwarder:

            time.sleep(3)

            # tell the target to accept incoming instances again
            self.ob.modalities.configure(
                modality='orthanc-a',
                configuration=orthanc_a_config
            )

            # wait until the source is empty (= the forwarder has completed its job and deleted them)
            helpers.wait_until(lambda: len(self.oa.studies.get_all_ids()) == 0, timeout=3000)  # TODO: 30 s

            # check only the flair series has arrived on b
            self.assertEqual(1, len(self.ob.instances.get_all_ids()))
            forwarded_instance = self.ob.instances.get(self.ob.instances.get_all_ids()[0])
            self.assertEqual("MY", forwarded_instance.tags.get('InstitutionName'))
            self.assertEqual("1234", forwarded_instance.tags.get('OtherPatientIDs'))
            self.assertEqual("sT2W/FLAIR", forwarded_instance.series.main_dicom_tags.get('SeriesDescription'))

            # check it has been removed from Orthanc A
            self.assertEqual(0, len(self.oa.instances.get_all_ids()))

    def test_orthanc_cleaner_with_past_studies(self):
        self.oa.delete_all_content()

        # populate Orthanc with and old study...
        populator = OrthancTestDbPopulator(
            api_client=self.oa,
            studies_count=1,
            series_count=1,
            instances_count=1,
            from_study_date=datetime.date(1980, 12, 16),
            to_study_date=datetime.date(1983, 12, 16)
        )
        populator.execute()

        old_study_id = self.oa.studies.get_all_ids()[0]

        # ...and a recent study
        populator = OrthancTestDbPopulator(
            api_client=self.oa,
            studies_count=1,
            series_count=1,
            instances_count=1,
            from_study_date=datetime.date.today()-datetime.timedelta(days=5),
            to_study_date=datetime.date.today()
        )
        populator.execute()

        # apply a label to one of the 2 studies. This label is NOT in the csv file
        studies_ids = self.oa.studies.get_all_ids()
        self.oa.studies.add_label(studies_ids[0], "LABEL5")

        cleaner = OrthancCleaner(api_client=self.oa, execution_time=None, labels_file_path=here / "stimuli/labels.csv")
        cleaner.execute()

        # check that if no correct label found, studies are kept
        self.assertEqual(len(self.oa.studies.get_all_ids()), 2)


        self.oa.studies.add_label(studies_ids[0], "LABEL2")
        self.oa.studies.add_label(studies_ids[1], "LABEL1")

        cleaner.execute()

        # we would like to check that the recent study is still there and the old one is gone,
        # but given the fact that they have just been uploaded, they should be kept
        self.assertEqual(len(self.oa.studies.get_all_ids()), 2)

    def test_orthanc_cleaner_with_future_studies(self):
        self.oa.delete_all_content()

        # We are not able to trick Orthanc to modify the `LastUpdate` value
        # so let's create studies with dates in the future and
        # a negative retention period

        # populate Orthanc with and "old" future study...
        populator = OrthancTestDbPopulator(
            api_client=self.oa,
            studies_count=1,
            series_count=1,
            instances_count=1,
            from_study_date=datetime.date.today() + datetime.timedelta(weeks=3),
            to_study_date=datetime.date.today() + datetime.timedelta(weeks=4)
        )
        populator.execute()

        old_study_id = self.oa.studies.get_all_ids()[0]

        # ...and a "recent" future study
        populator = OrthancTestDbPopulator(
            api_client=self.oa,
            studies_count=1,
            series_count=1,
            instances_count=1,
            from_study_date=datetime.date.today() + datetime.timedelta(weeks=14),
            to_study_date=datetime.date.today() + datetime.timedelta(weeks=16)
        )
        populator.execute()

        studies_ids = self.oa.studies.get_all_ids()

        self.oa.studies.add_label(studies_ids[0], "LABEL3")
        self.oa.studies.add_label(studies_ids[1], "LABEL4")

        cleaner = OrthancCleaner(api_client=self.oa, execution_time=None,
                                 labels_file_path=here / "stimuli/labels.csv")

        cleaner.execute()

        # we would like to check that the recent study is still there and the old one is gone,
        self.assertEqual(len(self.oa.studies.get_all_ids()), 1)
        self.assertNotEqual(old_study_id, self.oa.studies.get_all_ids()[0])


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    unittest.main()




import queue
import threading
import time
import os
import logging
import argparse
import datetime
import multiprocessing
import random
import pydicom
import uuid
from orthanc_api_client import helpers
from .helpers.scheduler import Scheduler

from orthanc_api_client import OrthancApiClient
logger = logging.getLogger(__name__)


class Message:
    def __init__(self, dicom_id: str = None, orthanc_id: str = None, should_stop: bool = False):
        self.dicom_id = dicom_id
        self.orthanc_id = orthanc_id
        self.should_stop = should_stop


class PacsMigrator:
    """
    Migrates DICOM studies from a SOURCE modality (usually a PACS) to another DESTINATION modality (usually Orthanc).
    The migrator must be attached to an Orthanc which can be the SOURCE or the DESTINATION

    There are multiple use cases to use this class:
    - as the SOURCE
      MIGRATOR --> DESTINATION
      To work in this setup, you must provide:
      - the destination_modality (defined in the MIGRATOR config)
      - you may optionally delete the images from the source afterward

    - as a passive intermediate between the source and destination.  This is useful to transfer images between the source and destination without modifying them.
      SOURCE --> DESTINATION
         |
      MIGRATOR
      To work in this setup, you must provide:
      - the source_modality (Orthanc alias defined in the MIGRATOR config)
      - the destination_aet (no need to define it in the MIGRATOR config)
      - set destination_modality to None

    - to populate Orthanc from a remote modality (in this case, the MIGRATOR is the target)
      SOURCE --> MIGRATOR
      To work in this setup, you must provide:
      - the source_modality (Orthanc alias defined in the MIGRATOR config)
      - set destination_modality to None
      - set destination_aet to None

    """

    def __init__(self,
                 api_client: OrthancApiClient,
                 from_study_date: datetime.date,        # Start date
                 to_study_date: datetime.date,          # End date
                 source_modality: str = None,           # Source modality as configured in Orthanc (alias)
                 max_cfind_study_count: int = None,     # Known maximum amount of studies retrievable from the source modality at once
                 destination_modality: str = None,      # Destination modality as configured in Orthanc (alias)
                 destination_aet: str = None,           # Destination AET
                 delete_from_source: bool = False,      # once the data has been migrated, delete it from source (only vali
                 scheduler: Scheduler = None,
                 worker_threads_count: int = multiprocessing.cpu_count() - 1  # by default, use all CPUs but one for compression
                 ):

        if (destination_aet is not None and destination_modality is not None):
            raise ValueError("You cannot define destinationAet and destinationModality together")

        self._api_client = api_client
        self._source_modality = source_modality
        self._from_study_date = from_study_date
        self._to_study_date = to_study_date
        self._max_cfind_study_count = max_cfind_study_count
        self._destination_modality = destination_modality
        self._destination_aet = destination_aet
        self._delete_from_source = delete_from_source
        self._scheduler = scheduler

        self._worker_threads_count = worker_threads_count
        self._worker_threads = []
        self._messages = queue.Queue(maxsize=2*worker_threads_count)  # this is thread safe https://docs.python.org/3.5/library/queue.html#module-queue
        self._is_running = False

        self._dicom_tags_to_query = {  # this might be extended once we implement filters
            'AccessionNumber': '',
            'PatientName': '',
            'StudyInstanceUID': ''
        }

        if not self._destination_modality and not self._destination_aet:
            # destination is orthanc -> set orthanc AET
            self._destination_aet = self._api_client.get_json('/system')["DicomAet"]

    @property
    def source_is_orthanc(self):
        return self._source_modality is None

    @property
    def target_is_orthanc(self):
        return self._destination_aet is None and self._destination_modality is None

    def process_messages(self, worker_thread_id: int):
        logger.debug(f"Starting Processing thread {worker_thread_id}")

        while True:
            message = self._messages.get()  # block until a message is available

            if message.should_stop:  # sent by stop() to stop all worker threads
                self._messages.task_done()
                break

            if self.source_is_orthanc:
                try:
                    logger.info(f"C-Store study {message.orthanc_id} from orthanc to destination modality {self._destination_modality}")
                    # move the study from orthanc to the target modality
                    self._api_client.modalities.send(
                        target_modality=self._destination_modality,
                        resources_ids=message.orthanc_id
                    )

                    if self._delete_from_source:
                        self._api_client.studies.delete(
                            orthanc_id=message.orthanc_id
                        )
                except Exception as ex:
                    logger.error(f"Error while transferring {message.orthanc_id} {str(ex)}")


            elif self._source_modality and self._destination_aet:
                retry_count = 0
                while retry_count < 5:
                    try:
                        logger.info(f"C-Move study {message.dicom_id} from source {self._source_modality} to destination AET {self._destination_aet}")
                        # move the study from source to target modality
                        self._api_client.modalities.move_study(
                            from_modality=self._source_modality,
                            dicom_id=message.dicom_id,
                            to_modality_aet=self._destination_aet
                        )
                        break
                    except Exception as ex:
                        retry_count += 1
                        if retry_count == 5:
                            logger.error(f"Error (retried 5 times) while transferring {message.dicom_id} {str(ex)}")
                        else:
                            logger.warning(f"Error while transferring, retrying... {message.dicom_id} {str(ex)}")

            else:
                raise NotImplementedError("configuration not handled")

            self._messages.task_done()  # tell the queue the item has been processed

        logger.debug(f"Processing thread {worker_thread_id} stopped")

    def push_message(self, message: Message):
        if self._scheduler:
            self._scheduler.wait_right_time_to_run()

        self._messages.put(message)


    def execute(self):
        if self._is_running:
            raise RuntimeError("Migrator is already running")

        if self._source_modality:
            logger.info("From Modality: " + self._source_modality)
        if self._destination_aet:
            logger.info("To AET: " + self._destination_aet)
        elif self._destination_modality:
            logger.info("To Modality: " + self._destination_modality)
        else:
            logger.info("To itself")

        logger.info("From Date: " + str(self._from_study_date))
        logger.info("To Date  : " + str(self._to_study_date))

        if self._scheduler:
            logger.info("Night & Week-end mode Enabled : " + str(self._scheduler._run_only_at_night_and_weekend))

        logger.info("Migrating with {n} threads".format(n = self._worker_threads_count))

        # create worker threads
        for i in range(0, self._worker_threads_count):
            self._worker_threads.append(threading.Thread(
                target=self.process_messages,
                name=f'Worker Thread {i}',
                args=(i,)
            ))

        # start threads
        self._is_running = True
        for worker_thread in self._worker_threads:
            worker_thread.start()

        if self._from_study_date < self._to_study_date:
            # ex 20220101 -> 20220131
            inc_date = 1
            to_date = self._to_study_date + datetime.timedelta(days=1)
        else:
            # ex 20220131 -> 20220101
            inc_date = -1
            to_date = self._to_study_date - datetime.timedelta(days=1)

        current_date = self._from_study_date

        while current_date != to_date:
            logger.info("Processing date {date}".format(date=str(current_date)))

            query = self._dicom_tags_to_query
            query["StudyDate"] = helpers.to_dicom_date(current_date)

            if self.source_is_orthanc:
                logger.info("Querying Orthanc")
                local_studies = self._api_client.studies.find(query=query)

                logger.info(f"Found {len(local_studies)} studies")

                for study in local_studies:
                    self.push_message(Message(orthanc_id=study.orthanc_id))
            else:
                logger.info(f"Querying remote modality {self._source_modality}")

                retry_count = 0
                while retry_count < 5:
                    try:
                        remote_modality_studies = self._api_client.modalities.query_studies(
                            from_modality=self._source_modality,
                            query=query
                        )
                        break

                    except Exception as ex:
                        retry_count += 1
                        if retry_count == 5:
                            logger.error("Could not query the modality (retried 5 times), aborting")
                            return

                logger.info(f"Found {len(remote_modality_studies)} studies")

                if self._max_cfind_study_count and len(remote_modality_studies) == self._max_cfind_study_count:
                    logger.error(f"Too many studies in a single request: {len(remote_modality_studies)}, you'll probably miss some studies")

                for study in remote_modality_studies:
                    self.push_message(Message(dicom_id=study.dicom_id))

            current_date += datetime.timedelta(days=inc_date)

        logger.info("Waiting for worker threads to complete")
        # post one 'empty' exit message per thread to unlock the threads from waiting on the process queue
        for i in range(0, self._worker_threads_count):
            self._messages.put(Message(should_stop=True))

        for worker_thread in self._worker_threads:
            worker_thread.join()

        self._is_running = False
        self._worker_threads = []

        logger.info("--------------------------------------------------------------------")
        logger.info("Migration completed")


# examples:
# migrator as source: MIGRATOR --> DESTINATION:  ! use the --delete_from_source option with care !!!
# python orthanc_tools/pacs_migrator.py --url=http://localhost:8042 --user=user --password=pwd --destination_modality=orthanc-debug --from_study_date=20000101 --to_study_date=20191231 --run_only_at_night_and_weekend --night_start_hour=18 --night_end_hour=6
#
#
# migrator aside from source and destination:
# SOURCE --> DESTINATION
#   |
# MIGRATOR
# python orthanc_tools/pacs_migrator.py --url=http://localhost:8044 --user=user --password=pwd --source_modality=service --destination_aet=ORTHANC --from_study_date=20000101 --to_study_date=20191231 --run_only_at_night_and_weekend --night_start_hour=18 --night_end_hour=6
#
#
# migrator as the destination: SOURCE --> MIGRATOR:
# python orthanc_tools/pacs_migrator.py --url=http://localhost:8044 --user=user --password=pwd --source_modality=service --destination_aet=ORTHANC-DEBUG --from_study_date=20000101 --to_study_date=20191231 --run_only_at_night_and_weekend --night_start_hour=18 --night_end_hour=6
#



if __name__ == '__main__':
    level = logging.INFO

    if os.environ.get('VERBOSE_ENABLED'):
        level = logging.DEBUG

    logging.basicConfig(level=level, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    parser = argparse.ArgumentParser(description='Migrate the content of a remote modality to another modality')
    parser.add_argument('--url', type=str, default=None, help='Orthanc url (migrator)')
    parser.add_argument('--user', type=str, default=None, help='Orthanc user name')
    parser.add_argument('--password', type=str, default=None, help='Orthanc password')
    parser.add_argument('--destination_modality', type=str, default=None, help='Destination modality (alias)')
    parser.add_argument('--destination_aet', type=str, default=None, help='Destination AET')
    parser.add_argument('--source_modality', type=str, default=None, help='Source modality (alias)')
    parser.add_argument('--from_study_date', type=str, required=True, help='From Study Date (format 20190225)')
    parser.add_argument('--to_study_date', type=str, required=True, help='To Study Date (format 20190225)')
    parser.add_argument('--delete_from_source', default=False, action='store_true', help='delete data from source (only if source is an Orthanc)')
    parser.add_argument('--worker_threads_count', type=int, default=1, help='Worker threads count')
    Scheduler.add_parser_arguments(parser)

    args = parser.parse_args()

    url = os.environ.get("ORTHANC_URL", args.url)
    user = os.environ.get("ORTHANC_USER", args.user)
    pwd = os.environ.get("ORTHANC_PWD", args.password)
    destination_modality = os.environ.get("DESTINATION_MODALITY", args.destination_modality)
    destination_aet = os.environ.get("DESTINATION_AET", args.destination_aet)
    source_modality = os.environ.get("SOURCE_MODALITY", args.source_modality)
    from_study_date = helpers.from_dicom_date(os.environ.get("FROM_STUDY_DATE", args.from_study_date))
    to_study_date = helpers.from_dicom_date(os.environ.get("TO_STUDY_DATE", args.to_study_date))
    worker_threads_count = int(os.environ.get("WORKER_THREADS_COUNT", str(args.worker_threads_count)))

    scheduler = Scheduler.create_from_args_and_env_var(args)

    if os.environ.get("DELETE_FROM_SOURCE", None) is not None:
        delete_from_source = os.environ.get("DELETE_FROM_SOURCE") == "true"
    else:
        delete_from_source = args.delete_from_source

    migrator = PacsMigrator(
        api_client=OrthancApiClient(url, user=user, pwd=pwd),
        from_study_date=from_study_date,
        to_study_date=to_study_date,
        destination_modality=destination_modality,
        destination_aet=destination_aet,
        source_modality=source_modality,
        delete_from_source=delete_from_source,
        scheduler=scheduler,
        worker_threads_count=worker_threads_count
    )

    migrator.execute()



from datetime import datetime
import random
import socketserver, subprocess, sys, re, socket
from threading import Thread
#from hl7Lib import Hl7MessageValidator, UnsupportedMessageType, InvalidHL7Message
from .hl7_message_validator import Hl7MessageValidator
from .hl7_error import UnsupportedMessageType, InvalidHL7Message
import hl7
import logging

logger = logging.getLogger(__name__)

#
# class UnsupportedMessageType(Exception):
#     """
#     Error that occurs when the :class:`MLLPServer` receives a message without an associated handler
#     """
#
#     def __init__(self, msgType):
#         self.msg_type = msgType
#
#     def __str__(self):
#         return 'No Handler found for message type %s' % self.msgType
#
#
# class InvalidHL7Message(Exception):
#     """
#     Error that occurs when the :class:`MLLPServer` receives a string which doesn't represent an ER7-encoded HL7 message
#     """
#
#     def __init__(self, msg):
#         self.msg = msg
#
#     def __str__(self):
#         return 'The string received is not a valid HL7 message : {0}'.format(self.msg)
#

class _Hl7MllpRequestHandler(socketserver.StreamRequestHandler):
    """
    Internal class: implements the request handler
    """

    def setup(self, encoding: str = 'iso-8859-1'):
        self.encoding = encoding
        self.sb = b'\x0b'
        self.eb = b'\x1c'
        self.cr = b'\x0d'
        self.timeout = self.server.timeout
        self._handlers = self.server._handlers
        self._validator = Hl7MessageValidator(encoding = encoding, sb = self.sb, eb = self.eb, cr = self.cr)

        socketserver.StreamRequestHandler.setup(self)

    def handle(self):
        end_seq = self.eb + self.cr  # end sequence for the MLLP message
        try:
            # receive the first 3 char of the byte stream (an MLLP HL7 message must at least be 3 chars long (the sb, eb and cr)
            bline = self.rfile.read(3)
        except socket.timeout:
            self.request.close()
            return

        if len(bline) < 3: # in case the client did not send a complete message
            return

        # check the first char of the MLLP message
        if bline[0] != self.sb[0]:
            self.request.close()
            return

        # receive chars one by one until we reach the end sequence
        while bline[-2:] != end_seq:
            try:
                char = self.rfile.read(1)
                if not char:
                    break
                bline += char
            except socket.timeout:
                self.request.close()
                return

        # convert the byte stream to string and try to extract a valid HL7 message out of it
        message = self._validator.validate(bline)
        if message is not None:
            try:
                response = self._route_message(message)
            except Exception as e:
                self.request.close()
            else:
                # encode the response
                if isinstance(response, hl7.Message):
                    self.wfile.write(self.sb + (str(response)).encode(self.encoding) + self.eb + self.cr)
                else:
                    self.wfile.write(self.sb + response + self.eb + self.cr)
        self.request.close()

    def _route_message(self, msg):
        try:
            try:
                msgType = msg.split('|')[8]  # message type is the 9th element of MSH segment (assumed to be the first segment)
            except Exception:
                raise InvalidHL7Message(msg = msg)

            try:
                handler, args = self._handlers[msgType][0], self._handlers[msgType][1:]
            except KeyError:
                raise UnsupportedMessageType(msgType)

            # create a new handler and call the constructor with the args provided
            return handler(msg, *args)
        except Exception as e:
            try:
                errHandler, args = self._handlers['ERR'][0], self._handlers['ERR'][1:]
            except KeyError:
                raise e
            else:
                response = errHandler(msg, error_description = str(e), *args)
                return response


def run_server_in_separate_thread(server):
    server.serve_forever()


class MLLPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    """
        A :class:`TCPServer <SocketServer.TCPServer>` subclass that implements an MLLP server.
        It receives MLLP-encoded HL7 and redirects them to the correct handler, according to the
        :attr:`handlers` dictionary passed in.

        The :attr:`handlers` dictionary is structured as follows. Every key represents a message type (i.e.,
        the MSH.9) to handle, and the associated value is a tuple containing a subclass of
        :class:`AbstractHandler` for that message type and additional arguments to pass to its
        constructor.

        It is possible to specify a special handler for errors using the ``ERR`` key.
        In this case the handler should subclass :class:`AbstractErrorHandler`,
        which receives, in addition to other parameters, the raised exception as the first argument.
        If the special handler is not specified the server will just close the connection.

        The class allows to specify the timeout to wait before closing the connection.

        :param host: the address of the listener (use '0.0.0.0' to accept connections from anywhere)
        :param port: the port of the listener
        :param handlers: the dictionary that specifies the handler classes for every kind of supported message.
        :param timeout: the timeout for the requests
    """
    allow_reuse_address = True

    def __init__(self, host, port, handlers, timeout = 10):
        self.host = host
        self.port = port
        self._handlers = {
            'ERR': (handle_error_message,)
            }
        self.add_handlers(handlers)

        self.timeout = timeout
        self.thread = None
        self._is_running = False
        super().__init__((host, port), _Hl7MllpRequestHandler)

    def add_handlers(self, handlers: dict):
        """
        Allow to add handler(s) to an existing server.
        No restart is needed.

        Use example:
        mllp_server = MLLPServer(
                    host = 'localhost',
                    port = 2575,
                    handlers = {
                    'ERR': (orm_handler.handle_error_message,)
                    },
                    logger = logging.getLogger('WORKLIST SERVER')
            )
        mllp_server.add_handler({'ORM^O01': (orm_handler.handle_orm_message,)})
        """
        self._handlers.update(handlers)

    def serve_forever(self):
        self._is_running = True
        logger.info("Starting MLLP Server listening on port {port}".format(port = self.port))
        socketserver.TCPServer.serve_forever(self)
        logger.info("Stopping MLLP Server")
        self._is_running = False

    def start(self):
        """ run the server in a separate thread
        call server.stop() from another thread to stop the server
        """
        self.thread = Thread(target = run_server_in_separate_thread, args = (self,))
        self.thread.start()

    def stop(self):
        """stops a server that has been started with start()"""

        if self.thread is not None:
            self.shutdown()
            self.thread.join()
            self.thread = None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def is_running(self):
        return self._is_running


def default_message_handler(message):
    return NotImplementedError("Please implement a message handler.")

def handle_error_message(self, message: str, error_description: str = None) -> hl7.Message:

    hl7_request = hl7.parse(message)  # we need to re-parse it here only the build the response

    hl7_response = hl7.parse('MSH|^~\&|{sending_application}||{receiving_application}|{receiving_facility}|{date_time}||ACK^O01|{ack_message_id}|P|2.3||||||8859/1\rMSA|AR|{message_id}|{error}'.format(  # TODO: handle encoding
        sending_application = hl7_request['MSH.F5.R1.C1'],
        receiving_application = hl7_request['MSH.F3.R1.C1'],
        receiving_facility = hl7_request['MSH.F4.R1.C1'],
        date_time = datetime.now().strftime("%Y%m%d%H%M%S"),
        message_id = hl7_request['MSH.F10.R1.C1'],
        ack_message_id = str(random.randrange(0, 10 ** 15)),
        error = error_description
    ))
    return hl7_response

# this is just a very quick usage example that does nothing usefull since it uses abstract handler
if __name__ == "__main__":
    # server = Hl7Server(port = 2575, requestHandler = Hl7BaseRequestHandler)
    server = MLLPServer('localhost', 2575, {
        'ORU^R01': (default_message_handler,),
        'ERR': (handle_error_message,)
    })
    # terminate with Ctrl-C
    try:
        server.serve_forever()
    except KeyboardInterrupt:
        sys.exit(0)
